__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import itertools
import random
import socket
import time

from kazoo.exceptions import NoNodeException
from uuid import uuid4

from samsa.config import ConsumerConfig
from samsa.consumer.partitions import PartitionOwnerRegistry
from samsa.exceptions import (SamsaException, PartitionOwnedError,
                              ImproperlyConfiguredError)

logger = logging.getLogger(__name__)


class Consumer(object):
    """Primary API for consuming kazoo messages as a group.
    """

    def __init__(self, cluster, topic, group):
        """
        :param cluster:
        :type cluster: :class:`samsa.cluster.Cluster`.
        :param topic: The topic to consume messages from.
        :type topic: :class:`samsa.topics.Topic`.
        :param group: The consumer group to join.
        :type group: str.

        """
        self.config = ConsumerConfig().build()
        self.cluster = cluster
        self.topic = topic
        self.group = group
        self.id = "%s:%s" % (socket.gethostname(), uuid4())

        self.id_path = '/consumers/%s/ids' % self.group

        self.partition_owner_registry = PartitionOwnerRegistry(
            self, cluster, topic, group)
        self.partitions = self.partition_owner_registry.get()

        path = '%s/%s' % (self.id_path, self.id)
        self.cluster.zookeeper.create(
            path, self.topic.name, ephemeral=True, makepath=True
        )

        self._rebalance()

    def _rebalance(self, event=None):
        """Joins a consumer group and claims partitions.

        """
        logger.info('Rebalancing consumer %s for topic %s.' % (
            self.id, self.topic.name)
        )

        zk = self.cluster.zookeeper
        broker_path = '/brokers/ids'
        try:
            zk.get_children(broker_path, watch=self._rebalance)
        except NoNodeException:
            raise ImproperlyConfiguredError(
                'The broker_path "%s" does not exist in your '
                'ZooKeeper cluster -- is your Kafka cluster running?'
                % broker_path)

        # 3. all consumers in the same group as Ci that consume topic T
        consumer_ids = zk.get_children(self.id_path, watch=self._rebalance)
        participants = []
        for id_ in consumer_ids:
            topic, stat = zk.get("%s/%s" % (self.id_path, id_))
            if topic == self.topic.name:
                participants.append(id_)
        # 5.
        participants.sort()

        # 6.
        i = participants.index(self.id)
        parts_per_consumer = len(self.topic.partitions) / len(participants)
        remainder_ppc = len(self.topic.partitions) % len(participants)

        start = parts_per_consumer * i + min(i, remainder_ppc)
        num_parts = parts_per_consumer + (0 if (i + 1 > remainder_ppc) else 1)

        # 7. assign partitions from i*N to (i+1)*N - 1 to consumer Ci
        new_partitions = itertools.islice(
            self.topic.partitions,
            start,
            start + num_parts
        )

        new_partitions = set(new_partitions)

        self.stop_partitions()

        # 8. remove current entries from the partition owner registry
        self.partition_owner_registry.remove(
            self.partitions - new_partitions
        )

        # 9. add newly assigned partitions to the partition owner registry
        for i in xrange(self.config['rebalance_retries_max']):
            try:
                # N.B. self.partitions will always reflect the most current
                # view of owned partitions. Therefor retrying this method
                # will progress.
                self.partition_owner_registry.add(
                    new_partitions - self.partitions
                )
                break
            except PartitionOwnedError, e:
                logger.debug("Someone still owns partition %s. Retrying"
                             % e)
                time.sleep(i ** 2)
                continue
        else:
            raise SamsaException("Couldn't acquire partitions.")

    def __iter__(self):
        """Iterate over available messages. Does not return.

        """
        while True:
            msg = self.next_message(self.config['consumer_timeout'])
            if not msg:
                time.sleep(1)
            else:
                yield msg

    def next_message(self, timeout=None):
        """Get the next message from one of the partitions.

        """
        return random.sample(self.partitions, 1)[0].next_message(timeout)

    def commit_offsets(self):
        """Commit the offsets of all messages consumed so far.

        """
        for partition in self.partitions:
            partition.commit_offset()

    def stop_partitions(self):
        """Stop partitions from fetching more threads.

        """
        self.commit_offsets()
        for partition in self.partitions:
            partition.stop()

    def empty(self):
        return all([p.empty() for p in self.partitions])



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random


def random_partitioner(partitions, key):
    """
    Returns a random partition out of all of the available partitions.
    """
    return random.choice(list(partitions))


class Partitioner(object):
    """
    The base class for custom class-based partitioners.
    """
    def __call__(self, partitions, key=None):
        raise NotImplementedError('Subclasses must define their own '
            ' partitioner implementation')


class HashingPartitioner(Partitioner):
    """
    Returns a (relatively) consistent partition out of all available partitions
    based on the key.

    Messages that are published with the same keys are not guaranteed to end up
    on the same broker if the number of brokers changes (due to the addition
    or removal of a broker, planned or unplanned) or if the number of topics
    per partition changes. This is also unreliable when not all brokers are
    aware of a topic, since the number of available partitions will be in flux
    until all brokers have accepted a write to that topic and have declared how
    many partitions that they are actually serving.

    :param hash_func: hash function (defaults to :func:`hash`), should return
        an `int`. If hash randomization (Python 2.7) is enabled, a custom
        hashing function should be defined that is consistent between
        interpreter restarts.
    :type hash_func: function
    """
    def __init__(self, hash_func=hash):
        self.hash_func = hash_func

    def __call__(self, partitions, key):
        """
        :param partitions: the partitions to choose
        :type partitions: sequence of partitions or
            :class:`samsa.partitions.PartitionMap`
        :param key: key used for routing
        :type key: any hashable type if using the default :func:`hash`
            implementation, any valid value for your custom hash function
        :returns: a partition
        :rtype: :class:`samsa.partitions.Partition`
        """
        if key is None:
            raise ValueError(
                'key cannot be `None` when using hashing partitioner'
            )
        partitions = list(partitions)
        return partitions[abs(self.hash_func(key)) % len(partitions)]


hashing_partitioner = HashingPartitioner()



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import mock
import unittest2
from kazoo.exceptions import NoNodeException

from samsa.cluster import Cluster
from samsa.exceptions import NoAvailablePartitionsError
from samsa.topics import TopicMap, Topic
from samsa.test.integration import KafkaIntegrationTestCase


class TopicIntgrationTestCase(KafkaIntegrationTestCase):
    def test_no_partitions(self):
        topic = self.kafka_cluster.topics['topic']

        self.kafka_broker.stop()

        with self.assertRaises(NoAvailablePartitionsError):
            topic.publish('message')


class TopicMapTest(unittest2.TestCase):
    def test_get_topic(self):
        topics = TopicMap(cluster=mock.Mock())
        topic = topics.get('topic-1')
        self.assertIsInstance(topic, Topic)

        # Retrieving the topic again should return the same object instance.
        self.assertIs(topic, topics.get('topic-1'))


class PartitionMapTest(unittest2.TestCase):
    def setUp(self):
        self.cluster = Cluster(zookeeper=mock.Mock())

    def test_configuration_no_node(self):
        def get(node, *args, **kwargs):
            if node.startswith('/brokers/ids'):
                return ('::', mock.Mock())
            else:
                raise NoNodeException

        def get_children(node, *args, **kwargs):
            if node.startswith('/brokers/ids'):
                return ['0', '1', '2']
            else:
                raise NoNodeException

        self.cluster.zookeeper.get = get
        self.cluster.zookeeper.get_children = get_children
        self.cluster.zookeeper.exists.return_value = None

        topic = self.cluster.topics.get('topic')
        self.assertEqual(len(topic.partitions), len(self.cluster.brokers))

    def test_configuration_with_nodes(self):
        nodes = {
            '0': '5',
            '1': '2',
            '3': '4',
        }

        def get_node_data(path):
            id = path.rsplit('/', 1)[-1]
            return (nodes[str(id)], mock.Mock())

        self.cluster.zookeeper.get_children.return_value = nodes.keys()
        self.cluster.zookeeper.get = get_node_data

        topic = self.cluster.topics.get('topic')
        self.assertEqual(len(topic.partitions), sum(map(int, nodes.values())))



import logging

from samsa.contrib.handler import KafkaHandler
from samsa.test.integration import KafkaIntegrationTestCase


class KafkaHandlerTestCase(KafkaIntegrationTestCase):
    def setUp(self, *args, **kwargs):
        super(KafkaHandlerTestCase, self).setUp(*args, **kwargs)
        self.topic = 'test'
        topic = self.kafka_cluster.topics[self.topic]

        self.logger = logging.getLogger('test')
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(KafkaHandler(topic))

    def test_emit(self):
        payload = 'Hello world!'
        self.logger.info(payload)

        def ensure_published():
            client = self.kafka_cluster.brokers[0].client
            messages = list(client.fetch(self.topic, 0, 0, 1000))
            self.assertEqual(len(messages), 1)
            self.assertIn(messages[0].payload, payload)

        self.assertPassesWithMultipleAttempts(ensure_published, 5)



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class SamsaException(Exception):
    pass


class ImproperlyConfiguredError(SamsaException):
    pass


class NoAvailablePartitionsError(SamsaException):
    pass


class PartitionOwnedError(SamsaException):
    pass


class InvalidVersionError(SamsaException):
    pass


# Protocol Client Exceptions

class SocketDisconnectedError(SamsaException):
    pass


class ProtocolClientError(SamsaException):
    # ERROR_CODE as specified by the protocol.
    ERROR_CODE = None


class UnknownError(ProtocolClientError):
    ERROR_CODE = -1


class OffsetOutOfRangeError(ProtocolClientError):
    ERROR_CODE = 1


class InvalidMessageError(ProtocolClientError):
    ERROR_CODE = 2


class WrongPartitionError(ProtocolClientError):
    ERROR_CODE = 3


class InvalidFetchSizeError(ProtocolClientError):
    ERROR_CODE = 4


ERROR_CODES = dict((exc.ERROR_CODE, exc) for exc in (
    UnknownError, OffsetOutOfRangeError, InvalidMessageError,
    WrongPartitionError, InvalidMessageError))



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import mock

from itertools import islice
from kazoo.testing import KazooTestCase

from samsa.test.integration import KafkaIntegrationTestCase, polling_timeout
from samsa.test.case import TestCase
from samsa.cluster import Cluster
from samsa.config import ConsumerConfig
from samsa.topics import Topic
from samsa.partitions import Partition
from samsa.consumer.partitions import PartitionOwnerRegistry, OwnedPartition


logger = logging.getLogger(__name__)


class TestPartitionOwnerRegistry(KazooTestCase):
    """Test the methods of :class:`samsa.consumer.PartitionOwnerRegistry`.
    """

    def setUp(self):
        super(TestPartitionOwnerRegistry, self).setUp()
        self.c = Cluster(self.client)
        self.c.brokers = mock.MagicMock()
        broker = mock.Mock()
        broker.id = 1
        self.c.brokers.__getitem__.return_value = broker

        self.consumer = mock.Mock()
        self.consumer.id  = 1234
        self.topic = mock.Mock()
        self.topic.name = 'topic'

        self.por = PartitionOwnerRegistry(
            self.consumer,
            self.c,
            self.topic,
            'group'
        )

        # Create 5 partitions with on the same topic and broker
        self.partitions = []
        for i in xrange(5):
            self.partitions.append(
                Partition(self.c, self.topic, broker, i)
            )

    def test_crd(self):
        """Test partition *c*reate, *r*ead, and *d*elete.
        """

        # Add the first 3 partitions to the registry and see that they're set
        self.por.add(self.partitions[:3])
        self.assertEquals(
            self.por.get(),
            set(self.partitions[:3])
        )

        # Remove the first partition and see that only [1, 2] exist
        self.por.remove([self.partitions[0]])
        self.assertEquals(
            self.por.get(),
            set(self.partitions[1:3])
        )

    def test_grows(self):
        """Test that the reference returned by
        :func:`samsa.consumer.partitions.PartitionOwnerRegistry.get` reflects
        the latest state.

        """
        partitions = self.por.get()
        self.assertEquals(len(partitions), 0)

        self.por.add(self.partitions)
        self.assertEquals(len(partitions), len(self.partitions))


class TestConsumer(KazooTestCase, TestCase):

    def setUp(self):
        super(TestConsumer, self).setUp()
        self.c = Cluster(self.client)

    def _register_fake_brokers(self, n=1):
        self.client.ensure_path("/brokers/ids")
        for i in xrange(n):
            path = "/brokers/ids/%d" % i
            data = "creator:127.0.0.1:%s" % (9092 + i)
            self.client.create(path, data)

    @mock.patch.object(OwnedPartition, '_fetch')
    def test_assigns_partitions(self, *args):
        """
        Test rebalance

        Adjust n_* to see how rebalancing performs.

        """
        n_partitions = 10
        n_consumers = 3
        self._register_fake_brokers(n_partitions)
        t = Topic(self.c, 'testtopic')

        consumers = [t.subscribe('group1') for i in xrange(n_consumers)]

        partitions = []
        for c in consumers:
            partitions.extend(c.partitions)

        # test that there are no duplicates.
        self.assertEquals(len(partitions), n_partitions)
        # test that every partitions is represented.
        self.assertEquals(len(set(partitions)), n_partitions)

    @mock.patch.object(Partition, 'fetch')
    def test_commits_offsets(self, fetch):
        """Test that message offsets are persisted to ZK.

        """
        self._register_fake_brokers(1)
        t = Topic(self.c, 'testtopic')

        c = t.subscribe('group')
        msgs = []
        for i in xrange(1, 10):
            msg = mock.Mock()
            msg.next_offset = 3 * i
            msg.payload = str(i) * 3
            msgs.append(msg)
        fetch.return_value = msgs

        self.assertPassesWithMultipleAttempts(
            lambda: self.assertTrue(c.next_message(10) is not None),
            5
        )
        self.assertEquals(len(c.partitions), 1)
        p = list(c.partitions)[0]

        self.assertEquals(p.offset, 3)
        c.commit_offsets()

        d, stat = self.client.get(p.path)
        self.assertEquals(d, '3')

    @mock.patch.object(Partition, 'fetch')
    def test_consumer_remembers_offset(self, fetch):
        """Test that offsets are successfully retrieved from zk.

        """
        topic = 'testtopic'
        group = 'testgroup'
        offset = 10

        fake_partition = mock.Mock()
        fake_partition.cluster = self.c
        fake_partition.topic.name = topic
        fake_partition.broker.id = 0
        fake_partition.number = 0
        fetch.return_value = ()

        op = OwnedPartition(fake_partition, group)
        op._current_offset = offset
        op.commit_offset()

        self._register_fake_brokers(1)
        t = Topic(self.c, topic)
        c = t.subscribe(group)

        self.assertEquals(len(c.partitions), 1)
        p = list(c.partitions)[0]
        self.assertEquals(p.offset, offset)

        self.assertEquals(None, p.next_message(0))
        fetch.assert_called_with(offset, ConsumerConfig.fetch_size)


class TestConsumerIntegration(KafkaIntegrationTestCase):

    def setUp(self):
        super(TestConsumerIntegration, self).setUp()
        self.kafka = self.kafka_broker.client

    def test_consumes(self):
        """Test that we can consume messages from kafka.

        """
        topic = 'topic'
        messages = ['hello world', 'foobar']

        # publish `messages` to `topic`
        self.kafka.produce(topic, 0, messages)

        t = Topic(self.kafka_cluster, topic)

        # subscribe to `topic`
        consumer = t.subscribe('group2')

        def test():
            """Test that `consumer` can see `messages`.

            catches exceptions so we can retry while we wait for kafka to
            coallesce.

            """
            logger.debug('Running `test`...')
            try:
                self.assertEquals(
                    list(islice(consumer, 0, len(messages))),
                    messages
                )
                return True
            except AssertionError as e:
                logger.exception('Caught exception: %s', e)
                return False

        # wait for one second for :func:`test` to return true or raise an error
        polling_timeout(test, 1)

        old_offset = [p.offset for p in consumer.partitions][0]
        # test that the offset of our 1 partition is not 0
        self.assertTrue(old_offset > 0)
        # and that consumer contains no more messages.
        self.assertTrue(consumer.empty())

        # repeat and see if offset grows.
        self.kafka.produce(topic, 0, messages)
        polling_timeout(test, 1)
        self.assertTrue([p.offset for p in consumer.partitions][0] > old_offset)


    def test_empty_topic(self):
        """Test that consuming an empty topic returns an empty list.

        """
        topic = 'topic'
        t = Topic(self.kafka_cluster, topic)

        consumer = t.subscribe('group2')
        self.assertTrue(consumer.empty())



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


def attribute_repr(*attributes):
    """
    Provides an alternative ``__repr__`` implementation that adds the values of
    the given attributes to the output as a development and debugging aid.

    For example::

        >>> class Foo(object):
        ...     __repr__ = attribute_repr('pk', 'slug')
        <foo.models.Foo at 0x100614c50: pk=1, slug=foo>

    :param \*attributes: a sequence of strings that will be used for attribute
        dereferencing on ``self``.
    """
    def _repr(self):
        cls = self.__class__
        pairs = ('%s=%s' % (attribute, repr(getattr(self, attribute, None)))
            for attribute in attributes)
        return u'<%s.%s at 0x%x: %s>' % (
            cls.__module__, cls.__name__, id(self), ', '.join(pairs))
    return _repr



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import atexit

from collections import namedtuple
from Queue import Queue
import threading


class ResponseFuture(object):
    """A samsa response which may have a value at some point."""

    def __init__(self, handler):
        self.handler = handler
        self.error = False
        self._ready = handler.Event()

    def set_response(self, response):
        """Set response data and trigger get method."""
        self.response = response
        self._ready.set()

    def set_error(self, error):
        """Set error and trigger get method."""
        self.error = error
        self._ready.set()

    def get(self, timeout=None):
        """Block until data is ready and return.

        Raises exception if there was an error.

        """
        self._ready.wait(timeout)
        if self.error:
            raise self.error
        return self.response


class Handler(object):

    def spawn(self, target, *args, **kwargs):
        raise NotImplementedError


# Could be possible to use the Kazoo handlers. Our RequestHandler is already
# implementation agnostic.
class ThreadingHandler(Handler):

    Queue = Queue
    Event = threading.Event

    def spawn(self, target, *args, **kwargs):
        t = threading.Thread(target=target, *args, **kwargs)
        t.daemon = True
        t.start()
        return t


class RequestHandler(object):
    """Uses a worker thread to dispatch requests."""

    Task = namedtuple('Task', ['request', 'future'])

    def __init__(self, handler, connection):
        self.handler = handler
        self.connection = connection
        self._requests = handler.Queue()
        self.ending = handler.Event()
        atexit.register(self.stop)

    def request(self, request, has_response=True):
        """Construct a new requst

        :returns: :class:`samsa.handlers.ResponseFuture`

        """
        future = None
        if has_response:
            future = ResponseFuture(self.handler)

        task = self.Task(request, future)
        self._requests.put(task)
        return future

    def start(self):
        """Start the request processor."""
        self.t = self._start_thread()

    def stop(self):
        """Stop the request processor."""
        self._requests.join()
        self.ending.set()

    def _start_thread(self):
        def worker():
            while not self.ending.is_set():
                task = self._requests.get()
                try:
                    self.connection.request(task.request)
                    if task.future:
                        self.connection.response(task.future)
                except Exception, e:
                    if task.future:
                        task.future.set_error(e)
                finally:
                    self._requests.task_done()

        return self.handler.spawn(worker)



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import collections
import functools
import itertools
import logging

from kazoo.exceptions import NoNodeException

from samsa.utils import attribute_repr
from samsa.utils.delayedconfig import (DelayedConfiguration,
    requires_configuration)


logger = logging.getLogger(__name__)


class PartitionMap(DelayedConfiguration):
    """
    Manages the partitions associated with a topic on a per-broker basis.

    :param cluster: The cluster that this partition map is associated with.
    :type cluster: :class:`samsa.cluster.Cluster`
    :param topic: The topic that this partition map is associated with.
    :type topic: :class:`samsa.topics.Topic`
    """
    def __init__(self, cluster, topic):
        self.cluster = cluster
        self.topic = topic

        self.__brokers = {}

    __repr__ = attribute_repr('topic')

    def _configure(self, event=None):
        path = '/brokers/topics/%s' % self.topic.name
        logger.info('Looking up brokers for %s...', self)

        try:
            broker_ids = self.cluster.zookeeper.get_children(
                path, watch=self._configure)
        except NoNodeException:
            if self.cluster.zookeeper.exists(path, watch=self._configure) \
                    is not None:
                self._configure()
            broker_ids = []

        brokers = map(self.cluster.brokers.get, map(int, broker_ids))

        # Add any broker IDs that are not already present in the mapping.
        for broker in brokers:
            if broker not in self.__brokers:
                partitionset = PartitionSet(self.cluster, self.topic, broker)
                logger.info('Discovered new partition set: %s', partitionset)
                self.__brokers[broker] = partitionset

        # Remove any brokers that are no longer present in the mapping.
        dead = set(self.__brokers.keys()) - set(brokers)
        for broker in dead:
            logger.info('Removing broker %s from %s', broker, self)
            del self.__brokers[broker]

    def __len__(self):
        """
        Returns the total number of partitions for this partition map,
        including virtual partitions.
        """
        partitions = list(iter(self))
        return len(partitions)

    def __iter__(self):
        """
        Returns an iterator containing every known partition for this topic.

        This includes "virtual" partitions for brokers that are present in the
        cluster, but have not yet registered their topic/partitions with the
        ZooKeeper cluster.
        """
        return itertools.chain(self.actual, self.virtual)

    @property
    @requires_configuration
    def actual(self):
        """
        Returns an iterator containing all of the partitions for this topic
        that have been registered by a broker in ZooKeeper.
        """
        return itertools.chain.from_iterable(itertools.imap(iter,
            self.__brokers.values()))

    @property
    @requires_configuration
    def virtual(self):
        """
        Returns an iterator containing "virtual" partitions for this topic.

        Virtual partitions are placeholder partitions for brokers that are
        known to be alive but are not aware of a topic. Since these brokers
        haven't seen the topic yet, they will not have published the number of
        partitions that they're serving. Every broker should be able to accept
        writes on the 0th partition for each topic, however, so the virtual
        partitions provide a partition objects for those partitions that have
        not yet been registered but are assumed to exist.
        """
        uninitialized_brokers = set(self.cluster.brokers.values()) - \
            set(self.__brokers.keys())
        create_virtual_partitionset = functools.partial(PartitionSet,
            cluster=self.cluster, topic=self.topic, virtual=True)
        virtual_iterator = lambda broker: \
            iter(create_virtual_partitionset(broker=broker))
        return itertools.chain.from_iterable(itertools.imap(virtual_iterator,
            uninitialized_brokers))


class PartitionSet(DelayedConfiguration):
    """
    Manages the partitions for a topic on a single broker.

    :param cluster: The cluster that this partition set is associated with.
    :type cluster: :class:`samsa.cluster.Cluster`
    :param topic: The topic that this partion set is associated with.
    :type topic: :class:`samsa.topics.Topic`
    :param broker: The broker this partition set is associated with.
    :type broker: :class:`samsa.brokers.Broker`
    :param virtual: Whether this is a "virtual" partition set or not. Virtual
        partition sets are used when a broker does not have any data associated
        with a specific topic.
    :type virtual: bool
    """
    def __init__(self, cluster, topic, broker, virtual=False):
        self.cluster = cluster
        self.topic = topic
        self.broker = broker
        self.virtual = virtual

        self.__count = None

    __repr__ = attribute_repr('topic', 'broker', 'virtual')

    def _configure(self, event=None):
        # TODO: At some point it might be wise to enable a "hint" size for
        # virtual partitions -- this way if brokers are serving a large number
        # of partitions, we can still maintain a reasonably even distribution
        # in a predictable environment. This might cause some problems in more
        # dynamic/less homogenous environments, however, since there's no good
        # way to track down a failure on a produce request to an invalid
        # partition.
        if self.virtual:
            count = 1
        else:
            node = '/brokers/topics/%s/%s' % (self.topic.name, self.broker.id)

            # If the node does not exist, this means this broker has not gotten
            # any writes for this partition yet. We can assume that the broker
            # is handling at least one partition for this topic, and update
            # when we have more information by setting an exists watch on the
            # node path.
            try:
                data, stat = self.cluster.zookeeper.get(node)
                count = int(data)
                logger.info('Found %s partitions for %s', count, self)
            except NoNodeException:
                if self.cluster.zookeeper.exists(node, watch=self._configure) \
                        is not None:
                    return self._configure()
                count = 1
                logger.info('%s is not registered in ZooKeeper, falling back '
                    'to %s virtual partition(s)', self, count)

        self.__count = count

    @requires_configuration
    def __len__(self):
        """
        Returns the total number of partitions available within this partition
        set.
        """
        return self.__count

    def __iter__(self):
        """
        Returns an iterator of :class:`samsa.topics.Partition` instances for
        this partition set.
        """
        for i in xrange(0, len(self)):
            yield Partition(self.cluster, self.topic, self.broker, i)


class Partition(object):
    def __init__(self, cluster, topic, broker, number):
        self.cluster = cluster
        self.topic = topic
        self.broker = broker
        self.number = int(number)

    __repr__ = attribute_repr('topic', 'broker', 'number')

    def publish(self, data):
        """
        Publishes one or more messages to this partition.
        """
        if isinstance(data, basestring):
            messages = [data]
        elif isinstance(data, collections.Sequence):
            messages = data
        else:
            raise TypeError

        return self.broker.client.produce(self.topic.name, self.number,
            messages)

    def fetch(self, offset, size):
        return self.broker.client.fetch(self.topic.name, self.number, offset,
            size)

    def __hash__(self):
        return hash((self.topic, self.broker.id, self.number))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self == other



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import unittest2

from samsa.utils.structuredio import StructuredBytesIO


class StructuredBytesIOTest(unittest2.TestCase):
    def test_length(self):
        buf = StructuredBytesIO('\x00\x00')
        self.assertEqual(len(buf), 2)

    def test_string_cast(self):
        value = '\x00\x00'
        buf = StructuredBytesIO(value)
        self.assertEqual(str(buf), value)
        buf.seek(2)  # SEEK_END
        self.assertEqual(str(buf), value)

    def test_write(self):
        buf = StructuredBytesIO()
        buf.write('test')
        self.assertEqual(str(buf), 'test')

        buf = StructuredBytesIO()
        buf.write(StructuredBytesIO('test'))
        self.assertEqual(str(buf), 'test')

    def test_pack(self):
        buf = StructuredBytesIO()
        buf.pack(2, 1)
        self.assertEqual(str(buf), '\x00\x01')

    def test_unpack(self):
        buf = StructuredBytesIO('\x00\x01')
        value = buf.unpack(2)
        self.assertEqual(value, 1)

    def test_frame(self):
        buf = StructuredBytesIO()
        buf.frame(1, 'test')
        self.assertEqual(str(buf), '\x04test')

    def test_unframe(self):
        buf = StructuredBytesIO('\x04test')
        self.assertEqual(buf.unframe(1), 'test')

        buf = StructuredBytesIO('\x04testextra')
        self.assertEqual(buf.unframe(1), 'test')

        buf = StructuredBytesIO('\x04tes')
        with self.assertRaises(ValueError):
            buf.unframe(1)

        buf = StructuredBytesIO('\x04tes')
        self.assertEqual(buf.unframe(1, validate=False), 'tes')

    def test_wrap(self):
        buf = StructuredBytesIO('test').wrap(1)
        self.assertIsInstance(buf, StructuredBytesIO)
        self.assertEqual(str(buf), '\x04test')

    def test_unwrap(self):
        buf = StructuredBytesIO('\x04test').unwrap(1)
        self.assertIsInstance(buf, StructuredBytesIO)
        self.assertEqual(str(buf), 'test')



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import mock
import unittest2
from kazoo.exceptions import NoNodeException

from samsa.brokers import BrokerMap, Broker
from samsa.cluster import Cluster
from samsa.exceptions import ImproperlyConfiguredError


class BrokerMapTest(unittest2.TestCase):
    def setUp(self):
        self.cluster = mock.Mock(spec=Cluster)
        self.cluster.zookeeper = mock.Mock()

    def test_configuration_no_node(self):
        self.cluster.zookeeper.get_children.side_effect = NoNodeException

        brokers = BrokerMap(self.cluster)
        with self.assertRaises(ImproperlyConfiguredError):
            brokers.get(0)

    def test_initial_configuration(self):
        nodes = ['0', '1', '2', '5']
        self.cluster.zookeeper.get_children.return_value = nodes

        brokers = BrokerMap(self.cluster)
        self.assertEqual(len(brokers), len(nodes))
        self.assertEqual(brokers.keys(), map(int, nodes))
        self.assertTrue(all(isinstance(value, Broker) for value
            in brokers.values()))

        self.assertEqual(self.cluster.zookeeper.get_children.call_count, 1)

    def test_update_configuration(self):
        nodes = ['0', '1']
        self.cluster.zookeeper.get_children.return_value = nodes
        brokers = BrokerMap(self.cluster)
        self.assertEqual(len(brokers), len(nodes))

        broker = brokers.get(1)

        # Emulate a broker entering the pool.
        nodes = ['0', '1', '2']
        self.cluster.zookeeper.get_children.return_value = nodes
        brokers._configure(event=mock.Mock())
        self.assertEqual(len(brokers), len(nodes))

        self.assertIs(broker, brokers.get(1))
        self.assertFalse(broker.is_dead)

        # Emulate a broker leaving the pool.
        nodes = ['0', '2']
        self.cluster.zookeeper.get_children.return_value = nodes
        brokers._configure(event=mock.Mock())
        self.assertEqual(len(brokers), len(nodes))

        self.assertTrue(broker.is_dead)
        with self.assertRaises(KeyError):
            brokers.get(1)


class BrokerTest(unittest2.TestCase):
    def setUp(self):
        self.cluster = mock.Mock(spec=Cluster)
        self.cluster.zookeeper = mock.Mock()

    def test_configuration(self):
        host = 'kafka-1.local'
        port = 9093
        template = '%(host)s-1342221875610:%(host)s:%(port)s'
        self.cluster.zookeeper.get.return_value = (template % {
            'host': host,
            'port': port,
        }, mock.Mock())

        broker = Broker(self.cluster, id_='1')
        self.assertEqual(broker.id, 1)
        self.assertEqual(self.cluster.zookeeper.get.call_count, 0)

        self.assertEqual(broker.host, host)
        self.assertEqual(broker.port, port)
        self.assertEqual(self.cluster.zookeeper.get.call_count, 1)



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from kazoo.exceptions import NodeExistsException, NoNodeException
from functools import partial
import Queue

from samsa.config import ConsumerConfig
from samsa.exceptions import PartitionOwnedError
from samsa.partitions import Partition


logger = logging.getLogger(__name__)


class OwnedPartition(Partition):

    """Represents a partition as a consumer group sees it.

    Manages offset tracking and message fetching.

    """

    def __init__(self, partition, group):
        """Initialize with a partition and consumer group.

        :param partition: Partition to initialize with.
        :type partition: :class:`samsa.partitions.Partition`.
        :param group: Group that owns this partition.
        :type group: str.

        """
        super(OwnedPartition, self).__init__(
            partition.cluster, partition.topic,
            partition.broker, partition.number
        )

        self.config = ConsumerConfig().build()

        self.group = group
        self.path = "/consumers/%s/offsets/%s/%s-%s" % (
            self.group, self.topic.name,
            self.broker.id, self.number
        )

        # _current_offset is cursor to next message we haven't consumed
        try:
            offset, stat = self.cluster.zookeeper.get(self.path)
            self._current_offset = int(offset)
        except NoNodeException:
            self.cluster.zookeeper.create(self.path, str(0), makepath=True)
            self._current_offset = 0

        # the offset at which we should make our next fetch
        self._next_offset = self._current_offset
        self._message_queue = Queue.Queue(self.config['queuedchunks_max'])
        self._fetch_thread = None

    @property
    def offset(self):
        return self._current_offset

    def empty(self):
        """True if there are no messages."""

        return self._message_queue.empty()

    def next_message(self, timeout=None):
        """Retrieve the next message for this partition.

        Returns None if no new messages and timeout elapses.

        :param timeout: block for timeout if integer, or indefinitely if None.

        """
        if not self._fetch_thread or not self._fetch_thread.is_alive():
            # TODO: turn this back into a long running thread if possible
            self._fetch_thread = self._create_thread()

        try:
            message = self._message_queue.get(True, timeout)
        except Queue.Empty:
            return None
        self._current_offset = message.next_offset
        return message.payload

    def commit_offset(self):
        """Commit current offset to zookeeper."""

        self.cluster.zookeeper.set(self.path, str(self._current_offset))

    def stop(self):
        """Stop the fetch thread."""

        if self._fetch_thread:
            self._fetch_thread.join()

    def _create_thread(self):
        return self.cluster.handler.spawn(
            target=self._fetch,
            args=(self.config['fetch_size'],)
        )

    def _fetch(self, size):
        """Fetch up to `size` bytes of new messages and add to queue.

        :param size: size in bytes of new messages to fetch.
        :type size: int

        """
        messages = super(OwnedPartition, self).fetch(
            self._next_offset,
            size
        )

        for message in messages:
            self._next_offset = message.next_offset
            logger.info('%s: Received message: %s', self, message)
            self._message_queue.put(
                message, True
            )

    def __del__(self):
        self.stop()


class PartitionOwnerRegistry(object):

    """Manages the Partition Owner Registry for a particular Consumer.
    """

    def __init__(self, consumer, cluster, topic, group):
        """
        :param consumer: consumer which owns these partitions.
        :type consumer: :class:`samsa.consumer.Consumer`.
        :param cluster:
        :type cluster: :class:`samsa.cluster.Cluster`.
        :param topic: The topic the partitions belong to.
        :type topic: :class:`samsa.topics.Topic`.
        :param group: The group the partitions belong to.
        :type group: str.

        """
        self.consumer_id = str(consumer.id)
        self.cluster = cluster
        self.topic = topic

        self.Partition = partial(OwnedPartition, group=group)
        self.path = '/consumers/%s/owners/%s' % (group, topic.name)
        self.cluster.zookeeper.ensure_path(self.path)
        self._partitions = set([])

    def get(self):
        """Get all owned partitions.

        """
        return self._partitions

    def remove(self, partitions):
        """Remove `partitions` from the registry.

        :param partitions: partitions to remove.
        :type partitions: iterable of :class:`samsa.partitions.Partition`.

        """
        for p in partitions:
            assert p in self._partitions
            self.cluster.zookeeper.delete(self._path_from_partition(p))
            self._partitions.remove(p)

    def add(self, partitions):
        """Add `partitions` to the registry.

        :param partitions: partitions to add.
        :type partitions: iterable of :class:`samsa.partitions.Partition`.

        """
        for p in partitions:
            try:
                self.cluster.zookeeper.create(
                    self._path_from_partition(p), self.consumer_id,
                    ephemeral=True
                )
            except NodeExistsException:
                raise PartitionOwnedError(p)
            self._partitions.add(self.Partition(p))

    def _path_from_partition(self, p):
        return "%s/%s-%s" % (self.path, p.broker.id, p.number)
